def external_force_potential(Ufunc):
    def force(time, position, orientation):
        pass

    return force

def external_torque_potential(Ufunc):
    def torque(time, position, orientation):
        pass

    return torque
